/***************************************************************************
 * MoioChart Library - graphs and charts made easy                         *
 * Copyright (C) 2005 by Silvio Moioli                                     *
 * silvio@moioli.net, http://www.moioli.net                                *  
 *                                                                         *
 * This program is free software; you can redistribute it and/or           *
 * modify it under the terms of the GNU General Public License             *
 * as published by the Free Software Foundation; either version 2          *
 * of the License, or (at your option) any later version.                  *
 *                                                                         *
 * This program is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with this program; if not, write to the Free Software             *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              *
 * MA  02110-1301, USA.                                                    *
 ***************************************************************************/

package net.moioli.drawables;

import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.lang.System;

import net.moioli.chart.GraphicsI;
import net.moioli.chart.Scale;

/**
 * A function which is defined (empirically) from a set of points, and
 * interpolated/ extrapolated with lines.
 * 
 * @author Silvio Moioli
 * @version 2005/10/5
 */
public class FunctionByPointsLines extends AbstractDrawableFunction implements
        FunctionByPointsI {
    /** Ascissae of the points */
    private double ascissae[] = null;

    /** Ordinate of the points */
    private double ordinate[] = null;

    /** Default constructor (y=x) */
    public FunctionByPointsLines() {
        super();
        double[] a = { 0, 1 };
        double[] o = { 0, 1 };
        this.ascissae = a;
        this.ordinate = o;
    }

    /** Sorts an array of DoublePoints by their x values. */
    private class AscissaeSorter implements Comparator {

        /**
         * Default constructor.
         */
        public AscissaeSorter() {
            super();
        }

        /**
         * Implements Comparator, takes two DoublePoints and compares their
         * ascissae.
         * 
         * @param o1
         *            first DoublePoint
         * @param o2
         *            second DoublePoint
         * @return a positive number if o1>o2, a negative one if o1 <o2 or zero
         *         if o1==o2.
         */
        public int compare(Object o1, Object o2) {
            return (int) (((DoublePoint) o1).getX() - ((DoublePoint) o2).getX());
        }
    }

    /**
     * Standard constructor.
     * 
     * @param points
     *            a DoublePoint Collection.
     */
    public FunctionByPointsLines(Collection points) {
        super();
        int size = points.size();

        double[] a = new double[size];
        double[] o = new double[size];

        DoublePoint[] puntiOrdinati = (DoublePoint[]) points
                .toArray(new DoublePoint[size]);
        Arrays.sort(puntiOrdinati, new AscissaeSorter());

        for (int i = 0; i < puntiOrdinati.length; i++) {
            a[i] = puntiOrdinati[i].getX();
            o[i] = puntiOrdinati[i].getY();
        }
        this.ascissae = a;
        this.ordinate = o;
    }

    /**
     * Returns the y value corresponding to the specified x.
     * 
     * @param x
     *            the abscissa
     * @return the ordinate
     */
    public double f(double x) {
        int search = Arrays.binarySearch(getAscissae(), x);
        double result;
        if (search < 0) {
            int ip = -(search + 1);
            double xP1 = 0;
            double yP1 = 0;
            double xP2 = 0;
            double yP2 = 0;
            if (ip > 0 && ip < getAscissae().length) {
                xP1 = getAscissae()[ip - 1];
                yP1 = getOrdinate()[ip - 1];
                xP2 = getAscissae()[ip];
                yP2 = getOrdinate()[ip];
            } else {
                if (ip == 0) {
                    xP1 = getAscissae()[0];
                    yP1 = getOrdinate()[0];
                    xP2 = getAscissae()[1];
                    yP2 = getOrdinate()[1];
                } else {
                    xP1 = getAscissae()[getAscissae().length - 1];
                    yP1 = getOrdinate()[getAscissae().length - 1];
                    xP2 = getAscissae()[getAscissae().length - 2];
                    yP2 = getOrdinate()[getAscissae().length - 2];
                }
            }
            double m = (yP2 - yP1) / (xP2 - xP1);
            double q = yP2 - m * xP2;
            result = x * m + q;
        } else {
            result = getOrdinate()[search];
        }
        return result;
    }

    /**
     * This method is automatically called by DefaultChart to draw the objects
     * it contains, so end users should never call drawMe directly. Inside this
     * method there must be all the code to draw the object on a DefaultChart,
     * and all the operations which involve scaling should be done via the
     * passed Scale object.
     * 
     * @param g
     *            the surface where this object will be drawn
     * @param s
     *            the DefaultChart's Scale object
     * @param c
     *            the color to be used for drawing
     */
    public void drawMe(GraphicsI g, Scale s, Object c) {
        g.setColor(c);
        if ((getAscissae().length > 2) && (getOrdinate().length > 2)) {

            // Extrapolates the first point
            double xP2 = getAscissae()[0];
            double yP2 = getOrdinate()[0];
            double xP3 = getAscissae()[1];
            double yP3 = getOrdinate()[1];
            double m = (yP3 - yP2) / (xP3 - xP2);
            double q = yP3 - m * xP3;

            double xP1 = s.getMinX();
            double yP1 = xP1 * m + q;

            // Draws the first segment
            g.drawLine(s.getX(xP1), s.getY(yP1), s.getX(xP2), s.getY(yP2));

            // Interpolates the middle points and draws them
            for (int i = 0; i < getAscissae().length - 1; i++) {
                double xPC = getAscissae()[i];
                double yPC = getOrdinate()[i];
                double xPN = getAscissae()[i + 1];
                double yPN = getOrdinate()[i + 1];
                g.drawLine(s.getX(xPC), s.getY(yPC), s.getX(xPN), s.getY(yPN));
            }

            // Estrapolates the last point
            double xPP = getAscissae()[getAscissae().length - 1];
            double yPP = getOrdinate()[getAscissae().length - 1];
            double xPT = getAscissae()[getAscissae().length - 2];
            double yPT = getOrdinate()[getAscissae().length - 2];
            m = (yPP - yPT) / (xPP - xPT);
            q = yPP - m * xPP;

            double xPL = s.getMaxX();
            double yPL = xPL * m + q;

            // and finally draws the last segment
            g.drawLine(s.getX(xPP), s.getY(yPP), s.getX(xPL), s.getY(yPL));
        }
    }

    /**
     * Returns the minimum y value that should be represented in the
     * DefaultChart for this object to be drawn properly. <br />
     * 
     * @param minX
     *            the minimum x value represented in this DefaultChart
     * @param maxX
     *            the maximum x value represented in this DefaultChart
     * @return the minimum y value
     */
    public double minY(double minX, double maxX) {
        double[] temp = new double[getOrdinate().length];
        System.arraycopy(getOrdinate(), 0, temp, 0, getOrdinate().length);
        Arrays.sort(temp);
        return temp[0];
    }

    /**
     * Returns the maximum y value that should be represented in the
     * DefaultChart for this object to be drawn properly. <br />
     * 
     * @param minX
     *            the minimum x value represented in this DefaultChart
     * @param maxX
     *            the maximum x value represented in this DefaultChart
     * @return the maximum y value
     */
    public double maxY(double minX, double maxX) {
        double[] temp = new double[getOrdinate().length];
        System.arraycopy(getOrdinate(), 0, temp, 0, getOrdinate().length);
        Arrays.sort(temp);
        return temp[temp.length - 1];
    }

    /**
     * Returns an array with the ascissae of the points that define this
     * function.
     * 
     * @return an array of double
     */
    public double[] getAscissae() {
        return this.ascissae;
    }

    /**
     * Changes the ascissae of the points that define this function.
     * 
     * @param ascissae
     *            the ascissae to set.
     */
    public void setAscissae(double[] ascissae) {
        this.ascissae = ascissae;
    }

    /**
     * Returns an array with the ordinate of the points that define this
     * function.
     * 
     * @return an array of double
     */
    public double[] getOrdinate() {
        return this.ordinate;
    }

    /**
     * Changes the ordinate of the points that define this function.
     * 
     * @param ordinate
     *            the ascissae to set.
     */
    public void setOrdinate(double[] ordinate) {
        this.ordinate = ordinate;
    }

}